/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* Copyright (C) 2010, Pascal Giorgi
 *
 * Written by Pascal Giorgi pascal.giorgi@lirmm.fr
 *
 * ========LICENCE========
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ========LICENCE========
 */


#ifndef __FFT_POLMUL__H
#define __FFT_POLMUL__H

#include <vector>
#include <cmath>
#include <complex.h>
#include <cassert>
#include <fftw3.h>

typedef std::complex<double>      Complex;
typedef std::complex<long double> LComplex;
struct fft_planner {
	size_t size;
	double *inA, *inB, *outC;
	//fftw_complex *outA, *outB, *inC;
	Complex *outA, *outB, *inC;
	fftw_plan pA,pB,pC;	

	fft_planner(){}
	
	fft_planner(size_t N){
		size = N;
		size_t N21=(N>>1)+1;
		inA  = (double*)      fftw_malloc(sizeof(double)       * N);
		inB  = (double*)      fftw_malloc(sizeof(double)       * N);
		outC = (double*)      fftw_malloc(sizeof(double)       * N);
		outA =(Complex*) fftw_malloc(sizeof(Complex) * N21);
		outB =(Complex*) fftw_malloc(sizeof(Complex) * N21);
		inC  =(Complex*) fftw_malloc(sizeof(Complex) * N21);		
		pA = fftw_plan_dft_r2c_1d(N, inA, reinterpret_cast<fftw_complex*>(outA), FFTW_MEASURE);
		pB = fftw_plan_dft_r2c_1d(N, inB, reinterpret_cast<fftw_complex*>(outB), FFTW_MEASURE);
		pC = fftw_plan_dft_c2r_1d(N, reinterpret_cast<fftw_complex*>(inC), outC, FFTW_MEASURE);
	}
	
	~fft_planner(){
		fftw_destroy_plan(pA);
		fftw_destroy_plan(pB);
		fftw_destroy_plan(pC);
		fftw_free(outA); 
		fftw_free(outB);
		fftw_free(outC);
		fftw_free(inA);
		fftw_free(inB);
		fftw_free(inC);		
	}

};

struct fft_launcher {
	
	static fft_planner* fft_plan;

	static void op_count(){
		double add,mul,fma;
		fftw_flops(fft_plan->pA,&add,&mul,&fma);
		std::cout<<"nbr op in planA: \n"<<mul<<" mul\n"<<add<<" add\n"<<fma<<" fma\n";
		fftw_print_plan(fft_plan->pA);
	}

	template<class T>
	static void FFT(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B){		
		size_t dA=A.size();
		size_t N=fft_plan->size, i;
		for (i=0;i<dA;i++){ fft_plan->inA[i]=A[i]; }		
		for(;i<N;i++)     { fft_plan->inA[i]=0.0 ; }		
		fftw_execute(fft_plan->pA); 
	}

	template<class T>
	static void fft_mul(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B){
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(fft_plan->size>=dC);
		
		size_t N=fft_plan->size, i;
		size_t N2=(N>>1)+1;
			
		for (i=0;i<dA;i++){
			fft_plan->inA[i]=A[i];
			fft_plan->inB[i]=B[i];
		}
		for(;i<N;i++){
			fft_plan->inA[i]=0.0;
			fft_plan->inB[i]=0.0;
		}
	
		fftw_execute(fft_plan->pA); 
		fftw_execute(fft_plan->pB);
#ifdef DEBUG
		for (i=0;i<N2;i++){std::cout<<"("<<creal(fft_plan->outA[i])<<"+I "<<cimag(fft_plan->outA[i])<<") ";}std::cout<<std::endl;
		for (i=0;i<N2;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") ";}std::cout<<std::endl;
#endif

		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif

		// scale the result
		double invN= 1./(double)N;
		for (i=0;i<dC;i++)
			C[i]=fft_plan->outC[i]*invN; 	       
	}


	// multiplication of Polynomial in Chebyshev Basis C=A*B using monomial basis reduction with fft 
	template<class T>
	static void cheby_mul_FFT(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B)
	{
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(fft_plan->size>=dC);
		
		size_t N=fft_plan->size, i;
		size_t N2=(N>>1)+1;


		/* compute A*B using FFT */
		for (i=0;i<dA;i++){
			fft_plan->inA[i]=A[i];
			fft_plan->inB[i]=B[i];
		}
		for(;i<N;i++){
			fft_plan->inA[i]=0.0;
			fft_plan->inB[i]=0.0;
		}
	
		fftw_execute(fft_plan->pA); 
		fftw_execute(fft_plan->pB);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outA[i])<<"+I "<<cimag(fft_plan->outA[i])<<") ";}std::cout<<std::endl;
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") ";}std::cout<<std::endl;
#endif	
		


		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif

		// scale the result by 1/2N (FFT 1/N + Cheby red 1/2)
		double invN= 1./(double)(N<<1);
		for (i=0;i<dC;i++)
			C[i]=fft_plan->outC[i]*invN; 

		/* compute A*Mirror(B) using FFT */
		// set A0 and B0 to zero		
		for (i=0;i<N2;i++){
			fft_plan->outA[i]-= fft_plan->inA[0];
		}
		for (i=0;i<dA;i++){
			fft_plan->inB[i]=B[dA-i-1];
		}
		fft_plan->inB[dA-1]=0.0;

		fftw_execute(fft_plan->pB);


		//std::cout<<"("<<creal(fft_plan->outB[N/3])<<"+I "<<cimag(fft_plan->outB[N/3])<<") \n"<<endl;

		
#ifdef DEBUG
		for (i=0;i<N2;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") \n";}std::cout<<std::endl;
#endif

		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);

		// scale the result
		invN= 1./(double)N;
		for (i=0;i<dC;i++)
			fft_plan->outC[i]*=invN; 
		
		// correct the result C by the result of A*Mirror(B) 
		C[0]+=fft_plan->outC[dA-1];//-A[0]*B[0];
		for (size_t i=1;i<dA-1;i++)
			C[i]+= 0.5*(fft_plan->outC[dA-1+i]+fft_plan->outC[dA-1-i]);//-A[0]*B[i] -A[i]*B[0]);

	}


	// multiplication of Polynomial in Chebyshev Basis C=A*B using monomial basis reduction with fft 
	template<class T>
	static void cheby_mul_FFT_V2(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B)
	{
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(fft_plan->size>=dC);
		
		size_t N=fft_plan->size, i;
		size_t N2=(N>>1)+1;
		size_t N4=(N>>2);

		/* compute A*B using FFT */
		for (i=0;i<dA;i++){
			fft_plan->inA[i]=A[i];
			fft_plan->inB[i]=B[i];
		}
		for(;i<N;i++){
			fft_plan->inA[i]=0.0;
			fft_plan->inB[i]=0.0;
		}
	
		fftw_execute(fft_plan->pA); 
		fftw_execute(fft_plan->pB);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outA[i])<<"+I "<<cimag(fft_plan->outA[i])<<") ";}std::cout<<std::endl;
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") ";}std::cout<<std::endl;
#endif	
		
		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif

		// scale the result by 1/2N (FFT 1/N + Cheby red 1/2)
		double invN= 1./(double)(N<<1);
		for (i=0;i<dC;i++)
			C[i]=fft_plan->outC[i]*invN; 

		/* compute A*Mirror(B) using FFT */
		// set A0 and B0 to zero		
		for (i=0;i<N2;i++){
			fft_plan->outA[i]-= fft_plan->inA[0];
			fft_plan->outB[i]-= fft_plan->inB[0];
		}
		

		/* compute the DFT of mirror(B) from DFT(B) */
		
		// switch coeff from 1 to dB // (use hermitian symmetry)
		for (i=1;i<N2;i++)
			fft_plan->outB[i]=conj(fft_plan->outB[i]);
		
		// scale coefficient with w^(d*i)		
		LComplex wd =std::exp(LComplex(0,1)*LComplex(2.*(M_PI*(dB-1))/(double)N,0.));
		LComplex wdi =wd;
			//long double _Complex wd=cexp(-2.0*_Complex_I* (M_PI*(dB-1))/(double)N);
			//long double _Complex wdi=wd;
		long double tmp;
		//cout.precision(30);
		//cout<<"N="<<N<<endl<<"d="<<dB-1<<endl;
		
		for (i=1;i<N2-1;i++){
			//std::cout<<"("<<creal(wdi)<<"+I "<<cimag(wdi)<<") "<<std::endl;		
			//wdi=cexp(-2.0*i*_Complex_I* (M_PI*(dB-1))/(double)N);
			fft_plan->outB[i]*=wdi;
			wdi*=wd;	
		}
		fft_plan->outB[N2-1]=Complex(0,0)-(fft_plan->outB[N2-1]);
		
		/*
		  for (i=1;i<N4;i++){			
		  wdi=cexp(-1.0*i*_Complex_I* (M_PI*(dB-1))/(double)(N>>1));
		  //std::cout<<"("<<creal(wdi)<<"+I "<<cimag(wdi)<<") "<<std::endl;		
		  //tmp=-2.0*i* (M_PI*(dB-1))/(double)N;
		  //wdi= cos(tmp)+_Complex_I*sin(tmp);
		  fft_plan->outB[i]     *=wdi;
		  fft_plan->outB[N2-1-i]*=(-creal(wdi)+_Complex_I*cimag(wdi));
		  //wdi*=wd;	
		  }
		  if (dB==2)
		  fft_plan->outB[N4]*= (0-_Complex_I) ;
		  else
		  fft_plan->outB[N4]*=_Complex_I ;

		  fft_plan->outB[N2-1]=0-(fft_plan->outB[N2-1]);
		*/
		//std::cout<<"("<<creal(fft_plan->outB[N/3])<<"+I "<<cimag(fft_plan->outB[N/3])<<") \n"<<endl;


#ifdef DEBUG
		for (i=0;i<N2;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") \n";}std::cout<<std::endl;
#endif

		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);

		// scale the result
		invN= 1./(double)N;
		fft_plan->outC[0]*=invN;
		invN*=0.5;
		for (i=1;i<dC;i++)
			fft_plan->outC[i]*=invN; 
		
		// correct the result C by the result of A*Mirror(B) 
		C[0]+=2.*fft_plan->outC[dA-1];//-A[0]*B[0];
		for (size_t i=1;i<dA-1;i++)
			C[i]+= (fft_plan->outC[dA-1+i]+fft_plan->outC[dA-1-i]);//-A[0]*B[i] -A[i]*B[0]);
	}

	// multiplication of Polynomial in Chebyshev Basis C=A*B using monomial basis reduction with fft 
	template<class T>
	static void cheby_mul_FFT_V3(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B)
	{
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(fft_plan->size>=dC);
		
		size_t N=fft_plan->size, i;
		size_t N2=(N>>1)+1;
		//size_t N4=(N>>2);

		/* compute A*B using FFT */
		for (i=0;i<dA;i++){
			fft_plan->inA[i]=A[i];
			fft_plan->inB[i]=B[i];
		}
		for(;i<N;i++){
			fft_plan->inA[i]=0.0;
			fft_plan->inB[i]=0.0;
		}
	
		fftw_execute(fft_plan->pA); 
		fftw_execute(fft_plan->pB);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outA[i])<<"+I "<<cimag(fft_plan->outA[i])<<") ";}std::cout<<std::endl;
		for (i=0;i<N;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") ";}std::cout<<std::endl;
#endif	
		
		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif

		// scale the result by 1/2N (FFT 1/N + Cheby red 1/2)
		double invN= 1./(double)(N<<1);
		for (i=0;i<dC;i++)
			C[i]=fft_plan->outC[i]*invN; 

		/* compute A*Mirror(B)*x using FFT */
		// set A0 and B0 to zero		
		for (i=0;i<N2;i++){
			fft_plan->outA[i]-= fft_plan->inA[0];
			fft_plan->outB[i]-= fft_plan->inB[0];
		}
		

		/* compute the DFT of mirror(B)*x from DFT(B) */
		
		// switch coeff from 1 to dB // (use hermitian symmetry)
		for (i=1;i<N2;i++)
			fft_plan->outB[i]=conj(fft_plan->outB[i]);
		
	
		// negate odd terms 
		for (i=1;i<N2-1;i=i+2)
			fft_plan->outB[i]=0. -fft_plan->outB[i];

#ifdef DEBUG
		for (i=0;i<N2;i++){std::cout<<"("<<creal(fft_plan->outB[i])<<"+I "<<cimag(fft_plan->outB[i])<<") \n";}std::cout<<std::endl;
#endif

		// pointwise multiplication of complex numbers (only half of them due to symmetry property used in fftw)
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i]*fft_plan->outB[i];
		  
		fftw_execute(fft_plan->pC);

		// NOTE THA outC holds A*Mirror(B)*x
		// each access to outC are thus shifted by 1 to get the correct value

		// scale the result
		invN= 1./(double)N;
		fft_plan->outC[0+1]*=invN;
		invN*=0.5;
		for (i=1;i<dC;i++)
			fft_plan->outC[i+1]*=invN; 
		
		// correct the result C by the result of A*Mirror(B) 
		C[0]+=2.*fft_plan->outC[dA];//-A[0]*B[0];
		for (size_t i=1;i<dA-1;i++)
			C[i]+= (fft_plan->outC[dA+i]+fft_plan->outC[dA-i]);//-A[0]*B[i] -A[i]*B[0]);
	}
};


struct dct_planner {
	size_t size;
	double  *inA,  *inB, *outC;
	double *outA, *outB, *inC;
	fftw_plan pA,pB,pC;	

	dct_planner(){}
	
	dct_planner(size_t N){
		size = N;
		//size_t N21=(N>>1)+1;
		inA  = (double*)      fftw_malloc(sizeof(double)       * size);
		inB  = (double*)      fftw_malloc(sizeof(double)       * size);
		inC  = (double*)      fftw_malloc(sizeof(double)       * size);
		outA = (double*)      fftw_malloc(sizeof(double)       * size);
		outB = (double*)      fftw_malloc(sizeof(double)       * size);
		outC = (double*)      fftw_malloc(sizeof(double)       * size);
		
		pA = fftw_plan_r2r_1d(size, inA,  outA, FFTW_REDFT00,  FFTW_MEASURE);
		pB = fftw_plan_r2r_1d(size, inB,  outB, FFTW_REDFT00,  FFTW_MEASURE);
		pC = fftw_plan_r2r_1d(size, inC,  outC, FFTW_REDFT00,  FFTW_MEASURE);
	}
	
	~dct_planner(){
		fftw_destroy_plan(pA);
		fftw_destroy_plan(pB);
		fftw_destroy_plan(pC);
		fftw_free(outA); 
		fftw_free(outB);
		fftw_free(outC);
		fftw_free(inA);
		fftw_free(inB);
		fftw_free(inC);		
	}

};

struct dct_launcher { 
	
	static dct_planner* dct_plan;

	static void op_count(){
		double add,mul,fma;
		fftw_flops(dct_plan->pA,&add,&mul,&fma);
		std::cout<<"nbr op in planA: \n"<<mul<<" mul\n"<<add<<" add\n"<<fma<<" fma\n";
		fftw_print_plan(dct_plan->pA);
	}

	template<class T>
	static void DCT(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B){
		size_t dA=A.size();
		size_t i, N=dct_plan->size;
		for (i=0;i<dA;i++){ dct_plan->inA[i]=A[i]; }		
		for(;i<N;i++)     { dct_plan->inA[i]=0.0 ; }		
		fftw_execute(dct_plan->pA); 
	}

	// multiplication of Polynomial in Chebyshev Basis C=A*B using DCT-I algorithm to get O(n log n) complexity
	template<class T>
	static void cheby_mul_DCT(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B)
	{
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(dC<=dct_plan->size);
	
		size_t i, N=dct_plan->size;

		for (i=0;i<dA;i++){ dct_plan->inA[i]=A[i]; dct_plan->inB[i]=B[i];}		
		for(;i<N;i++)     { dct_plan->inA[i]=0.0 ; dct_plan->inB[i]=0.0; }

		// DCT on inA and inB
		fftw_execute(dct_plan->pA); 
		fftw_execute(dct_plan->pB);

	
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<dct_plan->outA[i]<<" ";}std::cout<<std::endl;
		for (i=0;i<N;i++){std::cout<<dct_plan->outB[i]<<" ";}std::cout<<std::endl;
#endif
		// pointwise multiplication of double with scaling
		double scale= 0.25;//1.0/4.0;
		for (i=0;i<N;i++)
			dct_plan->inC[i]= dct_plan->outA[i]*dct_plan->outB[i]*scale;
		
		// DCT on inC
		fftw_execute(dct_plan->pC);
		
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<dct_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif
		// scale the result
		double hN=1.0/((double)N-1.0); 
		for (i=0;i<dC;i++)
			C[i]=dct_plan->outC[i]*hN; 		
	}
};




struct dct_via_fft_launcher{

	static fft_planner* fft_plan;

	// multiplication of Polynomial in Chebyshev Basis C=A*B using DCT-I algorithm to get O(n log n) complexity
	template<class T>
	static void cheby_mul_DCT(std::vector<T>& C, const std::vector<T>& A, const std::vector<T>& B)
	{
		size_t dA,dB,dC;
		dA=A.size();
		dB=B.size();
		dC=C.size();	
		assert(dA==dB);
		assert(dA+dB==dC+1);
		assert(dC<=(fft_plan->size)>>1);
	
		size_t i, N=fft_plan->size;
		size_t N2=(N>>1)+1;

		for (i=0;i<dA;i++){ fft_plan->inA[i]=A[i]; fft_plan->inB[i]=B[i];}		
		for(;i<N;i++)     { fft_plan->inA[i]=0.0 ; fft_plan->inB[i]=0.0; }
		fft_plan->inA[0]*=0.5;
		fft_plan->inB[0]*=0.5;

		// DCT on inA and inB
		fftw_execute(fft_plan->pA); 
		fftw_execute(fft_plan->pB);
	
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->inA[i]<<" ";}std::cout<<std::endl;
		for (i=0;i<N;i++){std::cout<<fft_plan->inB[i]<<" ";}std::cout<<std::endl;
#endif
		
		// pointwise multiplication of double with scaling
		double scale= 4.0;
		for (i=0;i<N2;i++)
			fft_plan->inC[i]= fft_plan->outA[i].real()*fft_plan->outB[i].real()*scale;				

		fftw_execute(fft_plan->pC);
#ifdef DEBUG
		for (i=0;i<N;i++){std::cout<<fft_plan->outC[i]<<" ";}std::cout<<std::endl;
#endif
		// scale the result
		double hN=1.0/(2.*(double)N); 
		for (i=0;i<dC;i++)
			C[i]=fft_plan->outC[i]*hN; 		
	}
};


#endif
